///////////////////////////////////////////////////////////////////////////////
//
// $Implemenatation:  MathObject.cpp$
//
// NAME: MathObject
//
// BASE CLASSES: none
//  
// PURPOSE: Implemenatation for class MathObject
//
// AUTHOR: Sidharth Thakur
//
// DATE: Thu Jun 30 14:44:04 EST 2005
//
// COPYRIGHT NOTICE: "Copyright 2005, Trustees of Indiana University, Hanson 
// Visualization Laboratory"
//
//  REVISION HISTORY:
// $Log:      $
///////////////////////////////////////////////////////////////////////////////

#include "MathObject.h"
#include "objectCurve.h"
#include "VisibleForm.h"

extern MathObject *mo2;
extern VisibleForm *parent;

MathObject::MathObject					( VisibleForm *parent ) 
{
#ifdef _DBG_
	////printf("inside MO::MathObject()\n");
#endif
    form = parent;
    objectID				= OBJCURVE;
	TORUS_PNTS				= 20;
	TORUS_STPS				= 20;
    flag_surf_init          = 1;
    nobjs                   = 0;
	initObject();
    globalVarInit();
	initGraph();
	updatePalette();
}

MathObject::~MathObject					( ) 
{
#ifdef _DBG_
	////printf("inside MO::~MathObject()\n");
#endif
    freeStructs();
}

void MathObject::freeStructs			( void ) 
{
#ifdef _DBG_
	////printf("inside MO::freeStructs()\n");
#endif
    int i = 0;
    int	j = 0;

    printf("CLEANUP DATA-STRUCTS... \n");
                      
    my_free(pntList);
    my_free(ctrlPntList);
    my_free(TangList);
    my_free(dTangList);
    my_free(basisRList);
    my_free(basisSList);
    my_free(refquatList);
    my_free(ptQuatList);
    my_free(fsQuatList);
    my_free(grQuatList);
	my_free(opQuatList);
    my_free(opupdQuatList);

    my_free(kfQuatList);

    my_free(ptRBasis);
    my_free(ptSBasis);
    my_free(fsRBasis);
    my_free(fsSBasis);
	my_free(grRBasis);
    my_free(grSBasis);

    my_free(qpntList);
    my_free(qTangList);
    my_free(qdTangList);
    my_free(CurvList);
    my_free(TorsnList);

    my_free(ptQuatList2);
    my_free(fsQuatList2);
    my_free(fsRBasis2);
    my_free(fsSBasis2);
    my_free(ptRBasis2);
    my_free(ptSBasis2);

    //tube ring structure
    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<radial_steps;j++)
            my_free(rpnts[i][j]);
    for(i=0;i<MAX_POINTS;i++)
        my_free(rpnts[i]);
    my_free(rpnts);

    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<radial_steps;j++)
            my_free(npnts[i][j]);
    for(i=0;i<MAX_POINTS;i++)
        my_free(npnts[i]);
    my_free(npnts);



    //quat-map tube structure
    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<radial_steps;j++)
            my_free(qrpnts[i][j]);
    for(i=0;i<MAX_POINTS;i++)
        my_free(qrpnts[i]);
    my_free(qrpnts);

    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<radial_steps;j++)
            my_free(qnpnts[i][j]);
    for(i=0;i<MAX_POINTS;i++)
        my_free(qnpnts[i]);
    my_free(qnpnts);

    //ring structure
    for(i=0;i<MAX_POINTS;i++)
        my_free(prevList[i]);
    my_free(prevList);

    for(i=0;i<MAX_POINTS;i++)
        my_free(currList[i]);
    my_free(currList);

    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<radial_steps;j++)
            my_free(rpList[i][j]);
    for(i=0;i<MAX_POINTS;i++)
        my_free(rpList[i]);
    my_free(rpList);

    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<radial_steps;j++)
            my_free(rpnList[i][j]);
    for(i=0;i<MAX_POINTS;i++)
        my_free(rpnList[i]);
    my_free(rpnList);


    //surface structure
    if(flag_surf_init != 1) {
        for(i=0;i<TORUS_PNTS;i++)
            for(j=0;j<TORUS_STPS;j++)
                my_free(surfGRQuatList[i][j]);
	    for(i=0;i<TORUS_PNTS;i++)
            my_free(surfGRQuatList[i]);
	    my_free(surfGRQuatList);

        for(i=0;i<TORUS_PNTS;i++)
            for(j=0;j<TORUS_STPS;j++)
                my_free(tpnts[i][j]);
	    for(i=0;i<TORUS_PNTS;i++)
            my_free(tpnts[i]);
	    my_free(tpnts);

        for(i=0;i<TORUS_PNTS;i++)
            for(j=0;j<TORUS_STPS;j++)
                my_free(upnts[i][j]);
	    for(i=0;i<TORUS_PNTS;i++)
            my_free(upnts[i]);
	    my_free(upnts);
    }

    //Line list data structures
    for(i=0;i<MAXOBJ;i++) 
        my_free(olist[i]);
    my_free(olist);

    if(tmpobj!=NULL) {
        for(i=0;i<mo2->nlines;i++) {
            my_free(tmpobj[i]);
        }
        my_free(tmpobj);
    }

}

void MathObject::initObject				( void ) 
{
#ifdef _DBG_
	//printf("inside MO::initObject()\n");
#endif
    setStatusBarText("Initializing application");
    if( objectID == OBJKNOT || objectID == OBJCURVE || objectID == OBJSURFACE ) {
        initStructs();
        initializeApplication( this );
        initCurve();
        whichCurve			= OBJSPLINE;
        flag_curve_closed	= 1;
    }
}

void MathObject::setStatusBarText       ( QString qstr ) 
{
    form->lb_status_text->setText(qstr);
}

void MathObject::initGraph				( void )
{
#ifdef _DBG_
	////printf("inside MO::initGraph()\n");
#endif
	gshowQdist		= 1;
	gshowQdotprod	= 1;
	gshowQtangacc	= 1;
	gshowCcurvature = 1;
	gshowCtorsion	= 1;
}
void MathObject::clearAll				( )
{
#ifdef _DBG_
	////printf("inside MO::clearAll()\n");
#endif
    initializeApplication( this );
	if( objectID == OBJCURVE) {
	    initCurve();
		initKnot();
	}
}
void MathObject::reinitStructs			( void )
{
#ifdef _DBG_
	////printf("inside MO::reinitStructs()\n");
#endif
	initializeApplication( this );
	initCurve();
}
void MathObject::initStructs			( void ) 
{
#ifdef _DBG_
	////printf("inside MO::initStructs()\n");
#endif
    int i = 0;
    int j = 0;
	//int	k = 0;
	
	//@ Global point list for 3D-curve
    pntList		= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS ); 

	//@ Global Control point list
    ctrlPntList	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS ); 

	//@ Global tangent list
    TangList	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS ); 

	//@ Global first-derivative-of-tangent list
    dTangList	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS ); 

	//@ Global frame-basis structs
    basisRList	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  
    basisSList	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  

    fsRBasis	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  
    fsSBasis	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  

    ptRBasis	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  
    ptSBasis	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  

    grRBasis	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  
    grSBasis	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  

	//@ Quaternion map structs
    refquatList	= (quaternion*) my_malloc ( sizeof(quaternion)*MAX_POINTS );

	//@ Quaternion map structs for PT Frame
    ptQuatList	= (quaternion*) my_malloc ( sizeof(quaternion)*MAX_POINTS );

	//@ Quaternion map structs for FS Frame
    fsQuatList	= (quaternion*) my_malloc ( sizeof(quaternion)*MAX_POINTS );

	//@ Quaternion map structs for FS Frame
    grQuatList	= (quaternion*) my_malloc ( sizeof(quaternion)*MAX_POINTS );

	opQuatList	= (quaternion*) my_malloc ( sizeof(quaternion)*MAX_POINTS );
	opupdQuatList = (quaternion*) my_malloc ( sizeof(quaternion)*MAX_POINTS );

	//@ optimization list
    kfQuatList = (quaternion*) my_malloc ( sizeof(quaternion)*MAX_POINTS );

	//@ Ring point structs
    radial_steps = 10;
    rpnts = (double***) my_malloc ( sizeof(double)*MAX_POINTS );
    for(i=0;i<MAX_POINTS;i++)
        rpnts[i] = (double**) my_malloc ( sizeof(double)*radial_steps );

    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<radial_steps;j++)
            rpnts[i][j] = (double*) my_malloc ( sizeof(double)*3 );

	//@ Normal points structs
    npnts = (double***) my_malloc ( sizeof(double)*MAX_POINTS );
    for(i=0;i<MAX_POINTS;i++)
        npnts[i] = (double**) my_malloc ( sizeof(double)*radial_steps );

    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<radial_steps;j++)
            npnts[i][j] = (double*) my_malloc ( sizeof(double)*3 );

	//# FOR QUATERNION MAP 
    qpntList	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS ); 
    qTangList	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS ); 
    qdTangList	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS ); 

	CurvList	= (double*) my_malloc ( sizeof(double)*MAX_POINTS);
	TorsnList	= (double*) my_malloc ( sizeof(double)*MAX_POINTS);

    ptQuatList2	= (quaternion*) my_malloc ( sizeof(quaternion)*MAX_POINTS );
    fsQuatList2	= (quaternion*) my_malloc ( sizeof(quaternion)*MAX_POINTS );
    fsRBasis2	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  
    fsSBasis2	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  
    ptRBasis2	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  
    ptSBasis2	= (Vertex*) my_malloc ( sizeof(Vertex)*MAX_POINTS );  

	//@ Ring point structs
    radial_steps = 10;
    qrpnts = (double***) my_malloc ( sizeof(double)*MAX_POINTS );
    for(i=0;i<MAX_POINTS;i++)
        qrpnts[i] = (double**) my_malloc ( sizeof(double)*radial_steps );

    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<radial_steps;j++)
            qrpnts[i][j] = (double*) my_malloc ( sizeof(double)*3 );

	//@ Normal points structs
    qnpnts = (double***) my_malloc ( sizeof(double)*MAX_POINTS );
    for(i=0;i<MAX_POINTS;i++)
        qnpnts[i] = (double**) my_malloc ( sizeof(double)*radial_steps );

    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<radial_steps;j++)
            qnpnts[i][j] = (double*) my_malloc ( sizeof(double)*3 );

	//Ring structs
	MAX_RING_PNTS = 300;

	prevList = (Vertex**) my_malloc(sizeof(Vertex)*500);
	for(i=0;i<MAX_RING_PNTS;i++)
		prevList[i] = (Vertex*) my_malloc(sizeof(Vertex));

	currList = (Vertex**) my_malloc(sizeof(Vertex)*MAX_RING_PNTS);
	for(i=0;i<MAX_RING_PNTS;i++)
		currList[i] = (Vertex*) my_malloc(sizeof(Vertex));

	rpList = (Vertex***) my_malloc(sizeof(Vertex)*MAX_POINTS);
	for(i=0;i<MAX_POINTS;i++)
		rpList[i] = (Vertex**) my_malloc(sizeof(Vertex)*MAX_RING_PNTS);
    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<MAX_RING_PNTS;j++)
			rpList[i][j] = (Vertex*) my_malloc(sizeof(Vertex));

	rpnList = (Vertex***) my_malloc(sizeof(Vertex)*MAX_POINTS);
	for(i=0;i<MAX_POINTS;i++)
		rpnList[i] = (Vertex**) my_malloc(sizeof(Vertex)*MAX_RING_PNTS);
    for(i=0;i<MAX_POINTS;i++)
        for(j=0;j<MAX_RING_PNTS;j++)
			rpnList[i][j] = (Vertex*) my_malloc(sizeof(Vertex));

}
void MathObject::initStructs_Surfaces   ( void )
{
    int i, j, k;

	tpnts		= (double***) my_malloc(sizeof(double)*MAX_POINTS);
	for(i=0;i<TORUS_PNTS;i++)
        tpnts[i] = (double**) my_malloc(sizeof(double)*TORUS_STPS);

    for(i=0;i<TORUS_PNTS;i++)
        for(j=0;j<TORUS_STPS;j++)
            tpnts[i][j] = (double*) my_malloc(sizeof(double)*3);

    for(i=0;i<TORUS_PNTS;i++)
        for(j=0;j<TORUS_STPS;j++)
            for(k=0;k<3;k++)
                tpnts[i][j][k]=0.0;

	upnts		= (double***) my_malloc(sizeof(double)*MAX_POINTS);
	for(i=0;i<TORUS_PNTS;i++)
        upnts[i] = (double**) my_malloc(sizeof(double)*TORUS_STPS);

    for(i=0;i<TORUS_PNTS;i++)
        for(j=0;j<TORUS_STPS;j++)
            upnts[i][j] = (double*) my_malloc(sizeof(double)*3);

    for(i=0;i<TORUS_PNTS;i++)
        for(j=0;j<TORUS_STPS;j++)
            for(k=0;k<3;k++)
                upnts[i][j][k]=0.0;

	surfGRQuatList = (quaternion***) my_malloc(sizeof(quaternion)*MAX_POINTS);
	for(i=0;i<TORUS_PNTS;i++)
        surfGRQuatList[i] = (quaternion**) my_malloc(sizeof(quaternion)*TORUS_STPS);

    for(i=0;i<TORUS_PNTS;i++)
        for(j=0;j<TORUS_STPS;j++)
            surfGRQuatList[i][j] = (quaternion*) my_malloc(sizeof(quaternion));

    for(i=0;i<TORUS_PNTS;i++)
		for(j=0;j<TORUS_STPS;j++) {
            surfGRQuatList[i][j]->r = 0.0;
			surfGRQuatList[i][j]->i = 0.0;
			surfGRQuatList[i][j]->j = 0.0;
			surfGRQuatList[i][j]->k = 0.0;
		}
}
void MathObject::computeObject			( void )
{
#ifdef _DBG_
	//////printf("inside MO::computeObject()\n");
#endif
    if(objectID == OBJCURVE && whichCurve == OBJKNOT) {
		////printf("inside for KNOT\n");
        constructKnot();
        makeTangent();
        makeTangentDerivatives();
        if(whichFrame == FRENETSERRET) {
            computeFrenetSerretFrame();
        }
        if(whichFrame == PARALLELTRANSPORT) {
            computeParallelTransportFrame();
        }
        if(whichFrame == GEODESICREFERENCE) {
            computeGeodesicReferenceFrame();
        }
        if(whichFrame != FRENETSERRET && selectFrame_FS) {
            computeFrenetSerretFrame();
        }
        if(whichFrame != PARALLELTRANSPORT && selectFrame_PT) {
            computeParallelTransportFrame();
        }
        if(whichFrame != GEODESICREFERENCE && selectFrame_GR) {
            computeGeodesicReferenceFrame();
        }
		computeCurveProperties();
		constructFramedCurve(tubeThickness);
		computeQuaternionRings();
		computeQuaternionRingsSurface_normals();
		computeQuaternionMap();
        //status.sprintf("%s %d,%d%s","Loaded",m,n,"-Torus Knot");
        //setStatusBarText(status);
	}//$ IF(OBJKNOT)
	if(objectID == OBJCURVE && whichCurve == OBJSPLINE) {
        computeSplineEquation();
        makeTangent();
        makeTangentDerivatives();

		//if(mo2->trackingRing == -1) {
			if(whichFrame == FRENETSERRET) {
				computeFrenetSerretFrame();
			}
			if(whichFrame == PARALLELTRANSPORT) {
				computeParallelTransportFrame();
			}
			if(whichFrame == GEODESICREFERENCE) {
				computeGeodesicReferenceFrame();
			}
			if(whichFrame == NOFRAME) {
			}


			if(whichFrame != FRENETSERRET && selectFrame_FS) {
				computeFrenetSerretFrame();
			}
			if(whichFrame != PARALLELTRANSPORT && selectFrame_PT) {
				computeParallelTransportFrame();
			}
			if(whichFrame != GEODESICREFERENCE && selectFrame_GR) {
				computeGeodesicReferenceFrame();
			}

			computeCurveProperties();
			constructFramedCurve(tubeThickness);
			//computeQuaternionRings();
			//computeQuaternionRingsSurface_normals();
			//computeQuaternionMap();

		//}//$ IF(OBJSPLINE)
	}
    if(objectID == OBJCURVE && whichCurve == OBJCURVELIST) {


    }
}
void MathObject::computeQuaternionMap	( void )
{
#ifdef _DBG_
	//////printf("inside MO::computeQuaternionMap()\n");
#endif
	if( objectID == OBJCURVE ) {
		if( selectFrame_FS ) {
			flag_quat_map_current_frame				= FRENETSERRET;
			qfgcolor[0] = 0.0f; qfgcolor[1] = 0.5f; qfgcolor[2] = 0.5f;
			if(qshowCurve)
				drawQuaternionCurve();
			copyQuatToPointList( FRENETSERRET );
			makeTangent_forQuatMap();
			makeTangentDerivatives_forQuatMap();
			computeParallelTransportFrame_forQuatMap();
		}
		if( selectFrame_PT ) {
			flag_quat_map_current_frame = PARALLELTRANSPORT;
			qfgcolor[0] = 0.0f; qfgcolor[1] = 0.5f; qfgcolor[2] = 0.0f;
			if(qshowCurve)
				drawQuaternionCurve();
			copyQuatToPointList( PARALLELTRANSPORT );
			makeTangent_forQuatMap();
			makeTangentDerivatives_forQuatMap();
			computeParallelTransportFrame_forQuatMap();
		}
		if( selectFrame_GR ) {
			flag_quat_map_current_frame = GEODESICREFERENCE;
			qfgcolor[0] = 0.5f; qfgcolor[1] = 0.0f; qfgcolor[2] = 0.0f;
			if(qshowCurve)
				drawQuaternionCurve();
			copyQuatToPointList( GEODESICREFERENCE );
			makeTangent_forQuatMap();
			makeTangentDerivatives_forQuatMap();
			computeParallelTransportFrame_forQuatMap();
		}
		if( flag_opt_only_once ) {
			initOPTIM();
			flag_opt_only_once = 0;
		}
	}
}

void MathObject::computeObject_forOpt   ( void )
{
	constructFramedCurve(tubeThickness);
    computeQuaternionMap();
}
void MathObject::display				( void ) 
{
#ifdef _DBG_
	//////printf("inside MO::display()\n");
#endif
 //   if(objectID == OBJCURVE && whichCurve == OBJSPLINE) {
	//	flag_curve_closed = 0;
 //       computeSplineEquation();
 //       makeTangent();
 //       makeTangentDerivatives();

	//	if(showLabels)
 //           drawLabels();
 //       if(showAxes)
 //           renderAxes();        
 //       if(showFrame) 
 //           drawFrames();
 //       if(showControlPoints)
 //           displayPoints( GL_RENDER );
 //       if(showCurve)
 //           drawCurve();
 //       if(showTSpace)
 //           drawTangentsDTangents();
	//	if(showFace || showEdge || showNormals) {
 //           renderFramedCurve(tubeThickness);
	//	}
	//	//drawCurveShadow();

	//}
    //else 
	if(objectID == OBJCURVE && whichCurve == OBJKNOT) {
        //constructKnot();
        //makeTangent();
        //makeTangentDerivatives();

        //if(whichFrame == FRENETSERRET) {
        //    computeFrenetSerretFrame();
        //}
        //if(whichFrame == PARALLELTRANSPORT) {
        //    computeParallelTransportFrame();
        //}
        //if(whichFrame == GEODESICREFERENCE) {
        //    computeGeodesicReferenceFrame();
        //}
        //if(whichFrame == NOFRAME) {
        //}


        //if(whichFrame != FRENETSERRET && selectFrame_FS) {
        //    computeFrenetSerretFrame();
        //}
        //if(whichFrame != PARALLELTRANSPORT && selectFrame_PT) {
        //    computeParallelTransportFrame();
        //}
        //if(whichFrame != GEODESICREFERENCE && selectFrame_GR) {
        //    computeGeodesicReferenceFrame();
        //}

		if(showLabels)
            drawLabels();
        if(showAxes)
            renderAxes();        
        if(showCurve && !showFace)
            drawKnottedCurve();
        if(showFrame) 
            drawFrames();
        if(showTSpace)
            drawTangentsDTangents();
        if(showFace || showEdge || showNormals)
            renderFramedCurve(tubeThickness);
	}
    if(objectID == OBJCURVE && whichCurve == OBJCURVELIST ) {

        if(mobjs>0)
            drawLINESET_tubed();
        if(flag_multi_lines && showCurve && !showFace) 
            drawLINESET();
        else if(!flag_multi_lines && showCurve)
            drawLINE();
        if(showFrame) 
            drawFrames();
        if(showTSpace)
            drawTangentsDTangents();

    }
	else if( objectID == OBJSURFACE ) {
        if(showAxes)
            renderAxes();
		switch( mo2->surfObjectID ) {
			case 0:
				computeTorus();
				break;
			case 1:
                computeSphere();
				break;
			case 2:
				computeCatenoid();
				break;
			case 3:
				computeEnneper();
				break;
			default:
				break;
		}		
		computeGeodesicReferenceFrame_MANIFOLD();
		renderSurfaceObject();
	}
}

void MathObject::displayPoints			( GLenum mode ) {
    drawPoints( mode );
}
void MathObject::showQuaternionMap		( void ) 
{

	if(objectID == OBJCURVE && whichCurve == OBJKNOT) {
		flag_quat_map_current_frame_for_QuatMap = PARALLELTRANSPORT;
		if(qshowAxes)
			renderAxes();
		if(qshowSphere)
			drawQuaternionUnitSphere();    
		if(qshowOptimPoints)
			drawQuaternionCurve_OPT();



		if( selectFrame_FS ) {
			flag_quat_map_current_frame	= FRENETSERRET;
			qfgcolor[0] = 0.0f; qfgcolor[1] = 0.1f; qfgcolor[2] = 0.8f;
			if(qshowCurve && !qshowFace)
				drawQuaternionCurve();
			copyQuatToPointList( FRENETSERRET );
			makeTangent_forQuatMap();
			makeTangentDerivatives_forQuatMap();
			computeParallelTransportFrame_forQuatMap();
			if(mo2->qshowFrame)
				drawFrames_forQuatMap();
            mo2->l_index_end = mo2->l_numPoints - 2;
			constructFramedCurve_forQuatMap(qmaptubeThickness);
			renderFramedQuaternionCurve_forFSMap ( );        
		}
		if( selectFrame_PT ) {
			flag_quat_map_current_frame = PARALLELTRANSPORT;
			qfgcolor[0] = 0.0f; qfgcolor[1] = 0.7f; qfgcolor[2] = 0.0f;
			if(qshowCurve && !qshowFace)
				drawQuaternionCurve();
			copyQuatToPointList( PARALLELTRANSPORT );
			makeTangent_forQuatMap();
			makeTangentDerivatives_forQuatMap();
			computeParallelTransportFrame_forQuatMap();
			if(mo2->qshowFrame)
				drawFrames_forQuatMap();
            mo2->l_index_end = mo2->l_numPoints - 3;
			constructFramedCurve_forQuatMap(qmaptubeThickness);
			renderFramedQuaternionCurve_forFSMap ( );        
		}
		if( selectFrame_GR ) {
			//printf("in _GR\n");
			flag_quat_map_current_frame = GEODESICREFERENCE;
			qfgcolor[0] = 0.5f; qfgcolor[1] = 0.0f; qfgcolor[2] = 0.0f;
			if(qshowCurve && !qshowFace)
				drawQuaternionCurve();
			copyQuatToPointList( GEODESICREFERENCE );
			makeTangent_forQuatMap();
			makeTangentDerivatives_forQuatMap();
			computeParallelTransportFrame_forQuatMap();
			if(mo2->qshowFrame)
				drawFrames_forQuatMap();
            mo2->l_index_end = mo2->l_numPoints - 3;
			constructFramedCurve_forQuatMap(qmaptubeThickness);
			renderFramedQuaternionCurve_forFSMap ( );        
		}

		if(qshowReflectedPoints) {

			copyQuatList();
			if(qshowCurve)
				drawQuaternionReflectedCurve();
			flag_quat_map_current_frame = REFLECTEDQUATLIST;
		    
			qfgcolor[0] = 0.5f; qfgcolor[1] = 0.0f; qfgcolor[2] = 0.5f;

			copyReflectedQuatToPointList();
		    
			makeTangent_forQuatMap();
			makeTangentDerivatives_forQuatMap();
			flag_quat_map_current_frame = whichFrame;
			if(mo2->qshowFrame) 
				drawFrames_forQuatMap();
		    
			computeParallelTransportFrame_forQuatMap();
			constructFramedCurve_forQuatMap(qmaptubeThickness);
			if(whichFrame == FRENETSERRET)
				renderFramedQuaternionCurve_forFSMap ( );
			else
				renderFramedCurve_forQuatMap(qmaptubeThickness);
		    
		}
		/*if( 1 ) {
			flag_quat_map_current_frame	= OPTIMIZATION;
			qfgcolor[0] = 0.8f; qfgcolor[1] = 0.6f; qfgcolor[2] = 0.1f;
			if(qshowCurve && !qshowFace)
				drawQuaternionCurve();
			copyQuatToPointList( OPTIMIZATION );
			makeTangent_forQuatMap();
			makeTangentDerivatives_forQuatMap();
			computeParallelTransportFrame_forQuatMap();
			if(mo2->qshowFrame)
				drawFrames_forQuatMap();
            mo2->l_index_end = mo2->l_numPoints - 2;
			constructFramedCurve_forQuatMap(qmaptubeThickness);
			renderFramedCurve_forQuatMap ( qmaptubeThickness );        
		}		*/
		drawQuaternionRings();
		if(flag_opt_first_time) {
			//initOPTIM();
			optCalcPTFrame_TurningAngle();
			flag_opt_first_time = 0;	
		}
		//optMinTurningAngle(opt_INDEX);
		//updateQuaternionRings();
		////test_FindQuat();
		////test_FindQuat( mo2->updatedTHETA );
		////if(mo2->whichCurve==OBJSPLINE && mo2->numCtrlPoints>3)
		//if(flag_opt_first_time) {
		//	printf("should print ONCE only!\n");
		//	initOPTIM();
		//	optMTAsetup();
		//	//copyQuatList_OPT();
		//	flag_opt_first_time = 0;
		//}
		//drawQuaternionCurve_OPT();
	}//$ IF(OBJCURVE..)
	if(objectID == OBJCURVE && whichCurve == OBJCURVELIST) {
		flag_quat_map_current_frame_for_QuatMap = PARALLELTRANSPORT;
        drawLINESET_QuatMap();
	}//$ IF(OBJCURVE..)
	//if(objectID == OBJCURVE && whichCurve == OBJSPLINE) {
	//	flag_quat_map_current_frame_for_QuatMap = PARALLELTRANSPORT;
	//	if(qshowAxes)
	//		renderAxes();
	//	if(qshowSphere)
	//		drawQuaternionUnitSphere();    
	//	drawQuaternionRings_forOpenCurve();
	//	if( selectFrame_FS ) {
	//		//printf("in _FS\n");
	//		flag_quat_map_current_frame	= FRENETSERRET;
	//		qfgcolor[0] = 0.0f; qfgcolor[1] = 0.1f; qfgcolor[2] = 0.8f;
	//		if(qshowCurve && !qshowFace)
	//			drawQuaternionCurve();
	//		copyQuatToPointList( FRENETSERRET );
	//		makeTangent_forQuatMap();
	//		makeTangentDerivatives_forQuatMap();
	//		computeParallelTransportFrame_forQuatMap();
	//		if(mo2->qshowFrame)
	//			drawFrames_forQuatMap();
	//			//needed here to draw all the FS,PT & GR maps simultaneously
	//			constructFramedCurve_forQuatMap(qmaptubeThickness);
	//			renderFramedQuaternionCurve_forFSMap ( );        
	//	}
	//	if( selectFrame_PT ) {
	//		//printf("in _PT\n");
	//		flag_quat_map_current_frame = PARALLELTRANSPORT;
	//		qfgcolor[0] = 0.0f; qfgcolor[1] = 0.3f; qfgcolor[2] = 0.0f;
	//		if(qshowCurve && !qshowFace)
	//			drawQuaternionCurve();
	//		copyQuatToPointList( PARALLELTRANSPORT );
	//		makeTangent_forQuatMap();
	//		makeTangentDerivatives_forQuatMap();
	//		computeParallelTransportFrame_forQuatMap();
	//		if(mo2->qshowFrame)
	//			drawFrames_forQuatMap();
	//			constructFramedCurve_forQuatMap(qmaptubeThickness);
	//			renderFramedQuaternionCurve_forFSMap ( );        
	//	}
	//	if( selectFrame_GR ) {
	//		//printf("in _GR\n");
	//		flag_quat_map_current_frame = GEODESICREFERENCE;
	//		qfgcolor[0] = 0.5f; qfgcolor[1] = 0.0f; qfgcolor[2] = 0.0f;
	//		if(qshowCurve && !qshowFace)
	//			drawQuaternionCurve();
	//		copyQuatToPointList( GEODESICREFERENCE );
	//		makeTangent_forQuatMap();
	//		makeTangentDerivatives_forQuatMap();
	//		computeParallelTransportFrame_forQuatMap();
	//		if(mo2->qshowFrame)
	//			drawFrames_forQuatMap();
	//			constructFramedCurve_forQuatMap(qmaptubeThickness);
	//			renderFramedQuaternionCurve_forFSMap ( );        
	//	}

	//	drawQuaternionRings();
	//	if(qshowReflectedPoints) {

	//		copyQuatList();
	//		if(qshowCurve)
	//			drawQuaternionReflectedCurve();
	//		flag_quat_map_current_frame = REFLECTEDQUATLIST;
	//	    
	//		qfgcolor[0] = 0.5f; qfgcolor[1] = 0.0f; qfgcolor[2] = 0.5f;

	//		copyReflectedQuatToPointList();
	//	    
	//		makeTangent_forQuatMap();
	//		makeTangentDerivatives_forQuatMap();
	//		flag_quat_map_current_frame = whichFrame;
	//		if(mo2->qshowFrame) 
	//			drawFrames_forQuatMap();
	//	    
	//		computeParallelTransportFrame_forQuatMap();
	//		constructFramedCurve_forQuatMap(qmaptubeThickness);
	//		if(whichFrame == FRENETSERRET)
	//			renderFramedQuaternionCurve_forFSMap ( );
	//		else
	//			renderFramedCurve_forQuatMap(qmaptubeThickness);
	//	    
	//	}
	//	//updateQuaternionRings();
	//	////test_FindQuat();
	//	////test_FindQuat( mo2->updatedTHETA );
	//	////if(mo2->whichCurve==OBJSPLINE && mo2->numCtrlPoints>3)
	//	//if(flag_opt_first_time) {
	//	//	printf("should print ONCE only!\n");
	//	//	initOPTIM();
	//	//	optMTAsetup();
	//	//	//copyQuatList_OPT();
	//	//	flag_opt_first_time = 0;
	//	//}
	//	//drawQuaternionCurve_OPT();
	//}//$ IF(OBJCURVE..)
	//else if( objectID == OBJSURFACE ) {
	//	selectFrame_FS = selectFrame_PT = 0;
	//	selectFrame_GR = 1;
	//	renderGaussMap();
	//}
 //   
}
void MathObject::showGaussMap			( void )
{
#ifdef _DBG_
	//printf("inside MO::showGaussMap()\n");
#endif
	if( objectID == OBJSURFACE ) {
		selectFrame_FS = selectFrame_PT = 0;
		selectFrame_GR = 1;
		drawQuaternionCurve_SURFACE();
	}
}
void MathObject::renderAxes				( void )
{
#ifdef _DBG_
	////printf("inside MO::renderAxes()\n");
#endif
    double len = 0.5;
    double rad = 0.025;
    
    GLUquadricObj *qobj;
    qobj = gluNewQuadric();

    glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT);
    glDisable(GL_LIGHTING);
	
    glPushMatrix();
         glTranslatef(0.0,0.0,0.0);
         glColor3f(1.0, 0.0, 0.0);
         glRotatef(-90,1,0,0);
         gluQuadricDrawStyle(qobj, GLU_FILL);
         gluCylinder(qobj, rad, rad, len, 16, 16);
    glPopMatrix();
    glPushMatrix();
         glTranslatef(0.0,0.0,0.0);
         glColor3f(0.0, 0.0, 1.0);
         glRotatef(90,0,1,0);
         gluQuadricDrawStyle(qobj, GLU_FILL);
         gluCylinder(qobj, rad, rad, len, 16, 16);
    glPopMatrix();
    glPushMatrix();
         glTranslatef(0.0,0.0,0.0);
         glColor3f(0.0, 1.0, 0.0);
         glRotatef(0,0,0,1);
         gluQuadricDrawStyle(qobj, GLU_FILL);
         gluCylinder(qobj, rad, rad, len, 16, 16);
    glPopMatrix();

    glPopAttrib();
    
}

void MathObject::testFunc				( void )
{
    TEST();
}

void MathObject::updateQuatRings		( void )
{
	optQuatCurveEnergy();
}
void MathObject::findQuaternion			( int i, quaternion *retQuat, double* retAngle )
{
	findQuat_Current( i, retQuat, (retAngle) );

}
void MathObject::initOPTIM				( void )
{
	//initOPTIM_STRUCT();
	curveQuatAngle_PT = 0.0;
	optCalcPTFrame_TurningAngle();
    initOPTIM_TurningAngle();
	//initOPTIM_TurningAngle();
    pt_init_value = 0.0;
    pt_update_ring_index = 2;
}
void MathObject::optimization			( void )
{
    //optCalcTurningAngle(opt_INDEX);
    //if(pt_flag_first_time) {
    //    optPT_step0();
    //    pt_flag_first_time = 0;
    //}
    //optPT_step0();
    //optPT_step(numPoints-1); //this is for optPT_step3 and below
    //printf("\t opt_INDEX = %d\n",opt_INDEX);
    optPT_step(opt_INDEX);
}
void MathObject::optPT_setLastQuat            ( void )
{
    optPT_step0();
}
void MathObject::optimization_INDEX		( int index ) {
	int l_index = index;
    l_index = 0;
    //optMTAstep(index);
}
void MathObject::computeCurveProperties ( void ) {
//#ifdef _DBG_
//	//printf("inside MO::computeCurveProperties()\n");
//#endif
	makeCurve_Curvature();
	makeCurve_Torsion();
	//initOPTIM();
    if(mo2->whichCurve != OBJCURVELIST) {
	    optMTAsetup();
	    for(int i = 1;i<numPoints-4; i++) {
		    computeProperties(i);
	    }
    }
}
/* Load Texture Files */
bool MathObject::texReadTextureFile		( const char *texFileName )
{	
	QString fileName = QString(texFileName);
	printf("Loading Texture file : %s...\n",texFileName );
	QFile FILE( fileName );
	if( !FILE.open( IO_ReadOnly ) ) {
		//ioError(  FILE, tr( "Cannot open DATA file %1 for writing!") );
		printf("Could not open texture file: %s \n",texFileName );
		return false;
	}
	FILE.close();
	texLoadTextureFile( texFileName );
	return true;
}
void MathObject::texLoadTextureFile		( const char *texFileName )
{
	char filename[80];
    const char* f = texFileName;
    
    strcpy(filename, f);    
    
    // 1. generate texture names
    glGenTextures ( 1, texName );
    
    // 2. bind textures
    glBindTexture ( GL_TEXTURE_2D, texName[0] );
    
    // 3. specify texture attributes
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,	    GL_CLAMP );
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,	    GL_CLAMP );
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
    glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    
    // 4. load image
    image = readPPM ( filename );
    flipVertical ( &image );
    
    // 5. enable texture
    // Note: Enable Texture only when required and where required
    if (image.buffer) {
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
      
        glBindTexture(GL_TEXTURE_2D, texName[0]); //create a texture object
        
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
        
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, image.width, image.height, 0, GL_RGB, GL_UNSIGNED_BYTE, image.buffer);
        
        free(image.buffer);
    }
    glDisable(GL_TEXTURE_2D);	
}
Image MathObject::readPPM				( const char *filename )
{
    Image image;
    FILE* fp;

    char header[3];
    char temp;

    int range;
    int size;
    //int ret;
    int ascii = 0;

    int i;
    int val;

    // Initialise image
    image.width = image.height = 0;
    image.buffer = NULL;
    
    // Read in the header.
    fp = fopen(filename,"rb");

    if (!fp) {
        printf("ERROR: Opening file!\n");
        return image;
    }

    if (fscanf(fp, "%2s\n", header)==1) {
		// Skip lines of comments beginning with #

        temp = (char) fgetc(fp);
        if (temp == '#')
            do 
                temp = (char) fgetc(fp);
            while (temp!='\n');
        else
            ungetc(temp,fp);
            
        // Read the width, height and range

        if (fscanf(fp, "%d %d\n%d\n", &image.width, &image.height, &range)==3) {
                //printf("height = %d, width = %d\n",image.height, image.width);
            
            // P6 binary RGB data
            // P4 ASCII RGB data
            if (strcmp(header,"P6")==0)
                ascii = 0;
            else
                if (strcmp(header,"P3")==0)
                    ascii = 1;
                else {
                    fprintf(stderr, "ASCII Header Error: texture %s is in incorrect format.\n", filename);
                    fclose(fp);
                    return image;
                } 

            if (range != 255) {
                fprintf(stderr, "Range Error: texture %s is in incorrect format.\n", filename);
                fclose(fp);
                return image;
            }

            // Allocate space for RGB texture
            size = image.width*image.height*3;
            image.buffer = (unsigned char *)calloc(size, sizeof(unsigned char)); 

            if (ascii)
                for (i=0; i<size; i++) {
                    fscanf(fp,"%d",&val);
                    image.buffer[i] = val;
                }
            else
                fread(image.buffer,sizeof(unsigned char),size,fp);
        }
    }
    fclose(fp);
    return image;
}

void MathObject::flipVertical			( Image *image )
{
    int i;
    unsigned char *temp,*src,*dest;
    int lineSize,totalSize;

    if (image==NULL)
        return;

    if (image->buffer==NULL)
        return;

    lineSize  = image->width*3;
    totalSize = image->height*lineSize;

    temp = (unsigned char *)calloc(totalSize, sizeof(unsigned char)); 

    dest = temp;
    src  = image->buffer + totalSize - lineSize;

    for (i=0; i<image->height; i++)
    {
        memcpy(dest,src,lineSize);
        dest += lineSize;
        src  -= lineSize;
    }

    free(image->buffer);
    image->buffer = temp;
}


/* Color Palette */
void MathObject::updatePalette          ( void )
{
    switch( mo2->whichPalette ) {

        case 0:
            load_palette_1();
            break;
        case 1:
            load_palette_2();
            break;
        case 2:
            load_palette_3();
            break;
        case 3:
            load_palette_4();
            break;
        default:
            break;
    }
}
void MathObject::load_palette_1			( void )
{
  int c[3];
  int i, j;
  double dc[3];
  int E0, DE, COLORSIZE;
  int C[6][3], D[5];

  COLORSIZE = 256;
  DE = COLORSIZE / 8;

  C[0][0] = 0; C[0][1] = 0; C[0][2] = 128; D[0] = DE;
  C[1][0] = 0; C[1][1] = 0; C[1][2] = 255; D[1] = 2*DE;
  C[2][0] = 0; C[2][1] = 255; C[2][2] = 255; D[2] = 2*DE;
  C[3][0] = 255; C[3][1] = 255; C[3][2] = 0; D[3] = 2*DE;
  C[4][0] = 255; C[4][1] = 0; C[4][2] = 0; D[4] = DE;
  C[5][0] = 128; C[5][1] = 0; C[5][2] = 0; 

  E0 = 0;
  for(j=0; j<5; j++) {
    dc[0] = (double) (C[j+1][0]-C[j][0]) / (double) D[j];
    dc[1] = (double) (C[j+1][1]-C[j][1]) / (double) D[j];
    dc[2] = (double) (C[j+1][2]-C[j][2]) / (double) D[j];
    for(i=E0; i<E0+D[j]; i++) {
      c[0] = C[j][0] + (int)((i-E0) * dc[0]);
      c[1] = C[j][1] + (int)((i-E0) * dc[1]);
      c[2] = C[j][2] + (int)((i-E0) * dc[2]);
      palette[i][0] = (float) c[0] / 255.0;
      palette[i][1] = (float) c[1] / 255.0;
      palette[i][2] = (float) c[2] / 255.0;
      palette[i][3] = 1.0;
    }
    E0 += D[j];
  }  
}
void MathObject::load_palette_2			( void )
{
  int c[3];
  int i, j;
  double dc[3];
  int E0, DE, COLORSIZE;
  int C[6][3], D[5];

  COLORSIZE = 256;
  DE = COLORSIZE / 8;

  C[0][0] = 0; C[0][1] = 0; C[0][2] = 128; D[0] = DE;
  C[1][0] = 0; C[1][1] = 0; C[1][2] = 255; D[1] = 2*DE;
  C[2][0] = 0; C[2][1] = 255; C[2][2] = 255; D[2] = 2*DE;
  C[3][0] = 255; C[3][1] = 255; C[3][2] = 0; D[3] = 2*DE;
  C[4][0] = 255; C[4][1] = 0; C[4][2] = 0; D[4] = DE;
  C[5][0] = 128; C[5][1] = 0; C[5][2] = 0; 

  E0 = 0;
  for(j=0; j<5; j++) {
    dc[0] = (double) (C[j+1][0]-C[j][0]) / (double) D[j];
    dc[1] = (double) (C[j+1][1]-C[j][1]) / (double) D[j];
    dc[2] = (double) (C[j+1][2]-C[j][2]) / (double) D[j];
    for(i=E0; i<E0+D[j]; i++) {
      c[0] = C[j][0] + (int)((i-E0) * dc[0]);
      c[1] = C[j][1] + (int)((i-E0) * dc[1]);
      c[2] = C[j][2] + (int)((i-E0) * dc[2]);
      palette[i][1] = (float) c[0] / 255.0;
      palette[i][2] = (float) c[1] / 255.0;
      palette[i][0] = (float) c[2] / 255.0;
      palette[i][3] = 1.0;
    }
    E0 += D[j];
  }  
}
void MathObject::load_palette_3			( void )
{
  int c[3];
  int i, j;
  double dc[3];
  int E0, DE, COLORSIZE;
  int C[6][3], D[5];

  COLORSIZE = 256;
  DE = COLORSIZE / 8;

  C[0][0] = 255; C[0][1] = 255; C[0][2] = 0; D[0] = DE;
  C[1][0] = 255; C[1][1] = 128; C[1][2] = 0; D[1] = 2*DE;
  C[2][0] = 255; C[2][1] = 64; C[2][2] = 0; D[2] = 2*DE;
  C[3][0] = 255; C[3][1] = 0; C[3][2] = 0; D[3] = 2*DE;
  C[4][0] = 128; C[4][1] = 0; C[4][2] = 0; D[4] = DE;
  C[5][0] = 0; C[5][1] = 0; C[5][2] = 0; 

  E0 = 0;
  for(j=0; j<5; j++) {
    dc[0] = (double) (C[j+1][0]-C[j][0]) / (double) D[j];
    dc[1] = (double) (C[j+1][1]-C[j][1]) / (double) D[j];
    dc[2] = (double) (C[j+1][2]-C[j][2]) / (double) D[j];
    for(i=E0; i<E0+D[j]; i++) {
      c[0] = C[j][0] + (int)((i-E0) * dc[0]);
      c[1] = C[j][1] + (int)((i-E0) * dc[1]);
      c[2] = C[j][2] + (int)((i-E0) * dc[2]);
      palette[i][0] = (float) c[0] / 255.0;
      palette[i][1] = (float) c[1] / 255.0;
      palette[i][2] = (float) c[2] / 255.0;
      palette[i][3] = 1.0;
    }
    E0 += D[j];
  }  
}
void MathObject::load_palette_4			( void )
{
  int c[3];
  int i, j;
  double dc[3];
  int E0, DE, COLORSIZE;
  int C[6][3], D[5];

  COLORSIZE = 256;
  DE = COLORSIZE / 8;

  C[0][0] = 0; C[0][1] = 255; C[0][2] = 255; D[0] = DE;
  C[1][0] = 0; C[1][1] = 128; C[1][2] = 255; D[1] = 2*DE;
  C[2][0] = 0; C[2][1] = 0; C[2][2] = 255; D[2] = 2*DE;
  C[3][0] = 0; C[3][1] = 0; C[3][2] = 128; D[3] = 2*DE;
  C[4][0] = 0; C[4][1] = 0; C[4][2] = 64; D[4] = DE;
  C[5][0] = 0; C[5][1] = 0; C[5][2] = 0; 

  E0 = 0;
  for(j=0; j<5; j++) {
    dc[0] = (double) (C[j+1][0]-C[j][0]) / (double) D[j];
    dc[1] = (double) (C[j+1][1]-C[j][1]) / (double) D[j];
    dc[2] = (double) (C[j+1][2]-C[j][2]) / (double) D[j];
    for(i=E0; i<E0+D[j]; i++) {
      c[0] = C[j][0] + (int)((i-E0) * dc[0]);
      c[1] = C[j][1] + (int)((i-E0) * dc[1]);
      c[2] = C[j][2] + (int)((i-E0) * dc[2]);
      palette[i][0] = (float) c[0] / 255.0;
      palette[i][1] = (float) c[1] / 255.0;
      palette[i][2] = (float) c[2] / 255.0;
      palette[i][3] = 1.0;
    }
    E0 += D[j];
  }  
}
/* Loading Meshview Data */
void MathObject::loadData               ( const char *FileName  ) 
{
    QString str;
    load_data(FileName );
    if(flag_multi_lines) {
        str.sprintf("File %s has been loaded with %d lines, %d vertices",FileName,lineList[0][0]->nlines,lineList[0][0]->nvertices);
        setStatusBarText(str);
    }
}
void MathObject::drawLINE               ( void )
{
	int k;
	Obj *obj;
	//Face *face;

	//glLineWidth(2);
	//if(redraw_flag == ON) {

		//glNewList(EDGE_LIST, GL_COMPILE_AND_EXECUTE);
		glDisable(GL_LIGHTING);
		//glColor4fv(black);
		//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		for(k=0; k<nobjs; k++) {
			obj = olist[k];
            //printf("\t nverts = %d, nfaces = %d\n",obj->nvertices, obj->nfaces);
            draw_line(obj, VERT_FRONT);
		}
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_LIGHTING);
		//glEndList();
	//} 
	//else {
	//	glCallList(EDGE_LIST);
	//}
}

void MathObject::drawLINESET            ( void )
{
	int k;
	Obj **obj;
	//Face *face;

	//glLineWidth(2);
	//if(redraw_flag == ON) {

		//glNewList(EDGE_LIST, GL_COMPILE_AND_EXECUTE);
		glDisable(GL_LIGHTING);
		//glColor4fv(black);
		//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        obj = mo2->lineList[0];
		for(k=0; k < nlines; k++) {
            draw_lines(obj[k], VERT_FRONT);
		}
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_LIGHTING);
		//glEndList();
	//} 
	//else {
	//	glCallList(EDGE_LIST);
	//}
}
void MathObject::drawLINESET_tubed      ( void )
{
    int i = 0;
    int inx = 0;
    mo2->numPoints = 0;
    mo2->radial_steps = 6;
    for(i=0; i<nlines; i++) {
        inx = i;
        mo2->numPoints = lineList[0][inx]->nvertices;
        copyLineList_toPointList(inx);
        makeTangent();
        makeTangentDerivatives();
		if(whichFrame == FRENETSERRET) {
			computeFrenetSerretFrame();
		}
		if(whichFrame == PARALLELTRANSPORT) {
			computeParallelTransportFrame();
		}
		if(whichFrame == GEODESICREFERENCE) {
			computeGeodesicReferenceFrame();
		}
		constructFramedCurve(0.02);
        computeCurveProperties();
        mo2->numPoints = lineList[0][inx]->nvertices-3;

        if(showFace || showEdge || showNormals)
            renderFramedCurve(tubeThickness);
    }
}
void MathObject::drawLINESET_QuatMap    ( void )
{
    int i = 0;
    int inx = 0;
    mo2->numPoints = 0;
    for(i=0; i<nlines; i++) {
        inx = i;
        mo2->numPoints = lineList[0][inx]->nvertices;
        copyLineList_toPointList(inx);
        makeTangent();
        makeTangentDerivatives();
		if(whichFrame == FRENETSERRET) {
			computeFrenetSerretFrame();
            qfgcolor[0] = 0.0f; qfgcolor[1] = 0.1f; qfgcolor[2] = 0.8f;
		}
		if(whichFrame == PARALLELTRANSPORT) {
			computeParallelTransportFrame();
            qfgcolor[0] = 0.0f; qfgcolor[1] = 0.7f; qfgcolor[2] = 0.0f;
		}
		if(whichFrame == GEODESICREFERENCE) {
			computeGeodesicReferenceFrame();
            qfgcolor[0] = 0.5f; qfgcolor[1] = 0.0f; qfgcolor[2] = 0.0f;
		}

        mo2->numPoints = lineList[0][inx]->nvertices-3;
        
        flag_quat_map_current_frame_for_QuatMap = PARALLELTRANSPORT;
        switch(whichFrame) {
            case FRENETSERRET:
                copyQuatToPointList( FRENETSERRET );
                break;
            case PARALLELTRANSPORT:
                copyQuatToPointList( PARALLELTRANSPORT );
                break;
            case GEODESICREFERENCE:
                copyQuatToPointList( GEODESICREFERENCE );
                break;
        }
		makeTangent_forQuatMap();
		makeTangentDerivatives_forQuatMap();

		computeParallelTransportFrame_forQuatMap();

        constructFramedCurve_forQuatMap(qmaptubeThickness);
		renderFramedCurve_forQuatMap (qmaptubeThickness );       

		if(mo2->qshowFrame)
			drawFrames_forQuatMap();
         if(qshowCurve && !qshowFace)
			drawQuaternionCurve();

    }
}
