#ifndef _MATHOBJECT_
#define _MATHOBJECT_

#include "all.h"

//#include "QuaternionSet.h"
#include <qdir.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qdatetime.h>
#include <qstring.h>

class VisibleForm;

class MathObject 
{
  public:
    MathObject( VisibleForm *parent );										// default constructor
    ~MathObject();										// default destructor

public:
    int			objectID;								// object ID: SPLINE,CURVE,KNOT,SURFACE
    QString     status;                                 // status bar text
    VisibleForm *form;

	//--------------------------------------------------//
	/* ADMIN ROUTINES */
    void		clearAll				( void );		// re-init App-state
    void		display					( void );		// main App-rendering routine
    void		displayPoints			( GLenum mode );// point-cloud rendering 
    void		initObject				( void );		// init object
    void		initStructs				( void );		// init data-structs
    void        initStructs_Surfaces    ( void );       // init surface data-structs
    void		freeStructs				( void );		// release data-structs
    void		reinitStructs			( void );
	void		renderAxes				( void );		// coord-axes rendering
    void		reset					( void );		// re-init App
    void        setStatusBarText        ( QString qstr );
    void		showQuaternionMap		( void );		// render Quat-Map
	void		showGaussMap			( void );
    void		testFunc				( void );		// function to test Quat-math routines
	
    void		updateQuatRings			( void );		// func to update quaternion rings
	void		findQuaternion			( int i, quaternion *retQuat, double* retAngle );

	void		optimization			( void );
	void		optimization_INDEX		( int index );
	void		initOPTIM				( void );
	void		optCalc					( void ); 

	//Computations
	void		computeObject			( void );
	void		computeQuaternionMap	( void );
    void        computeObject_forOpt    ( void );


	//Texture mgmt
    Image		readPPM					( const char *filename );
    void		flipVertical			( Image *image );
	bool		texReadTextureFile		( const char *texFileName );
	void		texLoadTextureFile		( const char *texFileName );
	Image		image;									// Id holder for texture
	GLuint		texName[MAX_TEXTURES];					// only one texture right now

	//Palette
	double		scale, obj_scale;
    float       palette[256][4];
	int			whichPalette;
    void        updatePalette           ( void );
	void		load_palette_1			( void );
	void		load_palette_2			( void );
    void		load_palette_3			( void );
    void		load_palette_4			( void );

	//--------------------------------------------------//
	/* # DATA POINTS */
    int			numPoints;								// # points for 3D-curve
	int			l_numPoints;
    int         l_index_end;
    int			radial_steps;							// # points tube-circumference
    int			numOfRings;								// # points on Quat-Map rings

	//--------------------------------------------------//
	/* # TEXTURE DATA */
	double		texParam;

	//--------------------------------------------------//
    /* DATA STRUCTURES */
	Vertex		*pntList;								// point-cloud for 3D-curve
    Vertex		*qpntList;								// point-cloud for Quat-curve
    Vertex		*ctrlPntList;							// point-cloud for editable 3D-curve (splines,lerp)

	int			MAX_RING_PNTS;
	int			numRingPoints;
	Vertex		**prevList;
	Vertex		**currList;
	Vertex		***rpList;								// list of Ring-points
	Vertex		***rpnList;								// list of Ring-surface normals

    double		***rpnts;								// list Curve-point vectors
	double		***npnts;								// list Curve-normal vectors

    double		***qrpnts;								// list Curve-point vectors
	double		***qnpnts;								// list Curve-normal vectors

	Vertex		*basisRList;							// list frame vector    	
    Vertex		*basisSList;							// list frame vector
        
    Vertex		*fsTList;								// list PT-frame vector
    Vertex		*fsRBasis;								// list PT-frame vector
    Vertex		*fsSBasis;								// list PT-frame vector

    Vertex		*ptTList;								// list FS-frame vector
    Vertex		*ptRBasis;								// list FS-frame vector
    Vertex		*ptSBasis;								// list FS-frame vector
        
    Vertex		*grTList;								// list GR-frame vector
    Vertex		*grRBasis;								// list GR-frame vector
    Vertex		*grSBasis;								// list GR-frame vector


    Vertex		*TangList;								// list Tangent vector
    Vertex		*dTangList;								// list Tangent-derivative vector
    Vertex		*qTangList;								// list Quat-Map curve Tangent vector
    Vertex		*qdTangList;							// list Quat-Map curve Tangent-derivative vector
        
	double		*CurvList;
	double		*TorsnList;

    Vertex		*ptRBasis2;								// list Quat-Map curve PT-frame vector
    Vertex		*ptSBasis2;								// list Quat-Map curve PT-frame vector

    Vertex		*fsRBasis2;								// list Quat-Map curve FS-frame vector
    Vertex		*fsSBasis2;								// list Quat-Map curve FS-frame vector

	double		***tpnts;								// list of torus points
	double		***upnts;								// list of torus normal points
    int			totalTorusPoints;						//
	int			TORUS_PNTS;								//
	int			TORUS_STPS;								//
	quaternion	***surfGRQuatList;
	
	//--------------------------------------------------//
	/* QUATERNION MAP STRUCTS */
    quaternion	*quatList;								// list Quat-points
	quaternion	*refquatList;							// list Quat-points: reflected
	
    quaternion	*ptQuatList;							// list Quat-points: PT-frame-map
    quaternion	*ptQuatList2;							// list Quat-points: PT-frame-map
	
    quaternion	*fsQuatList;							// list Quat-points: FS-frame-map
    quaternion	*fsQuatList2;							// list Quat-points: FS-frame-map

    quaternion	*grQuatList;							// list Quat-points: GR-frame-map 

	quaternion	*opQuatList,*opupdQuatList;

    int			flag_quat_map_current_frame;			// status flag: check current frame
    int			flag_quat_map_current_frame_for_QuatMap;// status flag: check current frame for Quat-Map
    int			flag_curve_closed;						//
    int			flag_quat_movingquat;					//
	int			flag_keypress_CTRL;
	int			flag_keypress_SHIFT;
	int			flag_keypress_ALT;

    double		*rthetaList;							// list Angles

	//--------------------------------------------------//
	/* STATUS ARRAY  */

    int			markFrame[MAX_POINTS];					// status array
    int			markFrameCounter;						// status array counter
    int			markQuatCurve[MAX_POINTS];				// status array
    int			markQuatCurveCounter;					// status array counter

	//--------------------------------------------------//
	/* DISPLAY OPTIONS */
	// > GLOBAL
    int			showAxes;								//
    int			showCurve;								//
    int			showControlPoints;						//
    int			showEdge;								//
    int			showFrame;								//
    int			showFace;								//
    int			showGap;								//
    int			showLabels;								//
    int			showMesh;								//
    int			showNormals;							//
	int			showOnScreenHelp;						//
	int			showPalette;
	int			showTexture;							//
	int			showColorCoded;							//
	int			showTSpace;								//
    int			whichSpline;							//
    int			whichCurve;								//
    int			whoCommandsMainWin;						//

    int     	showQuatMap;							//
    int     	showViewXYZ;							//
    int     	showViewWYZ;							//
    int			showViewWXY;							//
    int			showViewWXZ;							//
    int			whichFrame;								//
    int			selectFrame_FS;							//
    int			selectFrame_PT;							//
    int			selectFrame_GR;							//
    int			whichQuatFrame;							//
    int			whichQuatComponent;						//

	// > QUATERNION MAP 
    void		*qfont;									//
    int			qshowAxes;								//
    int			qshowCurve;								//
    int			qshowEdge;								//
    int			qshowFrame;								//
    int			qshowFace;								//
    int			qshowNormals;							//
	int			qshowOptimPoints;						//
    int			qshowPoints;							//
    int     	qshowQuatMap;							//
    int			qshowQuatRings;							//
    int			qshowQuatRingsSurface;					//
    int			qshowQuatRingsSurfaceMesh;				//
    int			qshowReflectedPoints;					//
    int			qshowSphere;							//
    int     	qshowViewXYZ;							//
    int     	qshowViewWYZ;							//
    int			qshowViewWXY;							//
    int			qshowViewWXZ;							//
    
	int			qwhichFrame;							//
    int			qwhichQuatComponent;					//
    int			qwhichQuatFrame;						//
    int			qwhichSpline;							//

	// > QUATERNION RING
    int			rshowAxes;								//
    int			rshowCurve;								//
    int			rshowEdge;								//
    int			rshowFace;								//
    int			rshowFrame;								//
	int			rshowSurfaceNormals;					//
	int			rshowSurfacePoints;						//
	double		rquatRingAngle;
	double		rtrackQuatAngle;
	//Ring texture
	int			rshowColorMode;
	//--------------------------------------------------//


	//--------------------------------------------------//
	/* GLOBAL MATERIAL OPTIONS */
    double		fgcolor[3];								//
    double		bgcolor[3];								//
    double		qfgcolor[3];							// for quaternion-maps
    double		qbgcolor[3];							//
	/* set light matrices (ambient, diffuse, specular) */
	float		mata[4],matd[4],mats[4];
	double		shininess;
	double		specular;

	//--------------------------------------------------//
	/* OBJECT CURVE OPTIONS */
    int			selected_point;
    int			numCtrlPoints;
    double		param_spline_increment;

	//--------------------------------------------------//
	/* OBJECT KNOT OPTIONS */
    int			m;										//
    int			n;										//
    double		r;										//
    double		a;										//
    double		b;										//
    int			lock_a_b;								//
    double		startTheta;								//

    double		increment;								// step
    double		tubeThickness;							//
    double      qmaptubeThickness;						//
    double		texparam;								//
    int			flag_end_rings_match;					//
    int			flag_update_track_ring;					//
    int			flag_track_ring;						//
    int			flag_edit_quatmap;						//
    int			trackingRing;							//
    int			trackingRingTheta;						//

	//--------------------------------------------------//
	/* OBJECT KNOT SURFACE OPTIONS */
	int			u;
	int			v;
	double		param_a;
	double		param_b;
	double		param_r;
	int			surfNumObjects;
	int			surfObjectID;
	
    //--------------------------------------------------//
    /* SURFACE DATA STRUCTURES */
    int         flag_surf_init;

	//--------------------------------------------------//
	/* OBJECT OPTIMIZATION OPTIONS */
	double		curveQuatAngle_PT;
	double		curveQuatDistance;
	double		curveQuatAngle;
	double		curveQuatDistance_CNT;
	double		curveQuatAngle_CNT;
	double		opt_EPSILON;
	double		opt_ERROR;
	int			opt_INDEX;
	double		updatedTHETA;
	int			flag_opt_first_time;
	int			flag_opt_only_once;
	int			flag_opt_continue;
	//MTA METHOD
		double		minTurningAngle;
		double		mtaENERGY;
		double		list_mtaENERGY_dist[MAX_POINTS];
		double		list_mtaENERGY_angle[MAX_POINTS];
		double		list_mtaENERGY_tacc[MAX_POINTS];
		double		list_ENERGY_turnang[MAX_POINTS];
		int			turnang_markIndex;
		int			limit_pmin;
		int			limit_pmax;
		int			list_keyframes[10];
		int			num_keyframes;
		int			num_total_frames;
		quaternion	*kfQuatList;

        quaternion	prevQuat, currQuat;
		double		curve_energy;
		int			index_q0;
		int			index_q1;
		int			index_q2;
		double		gmax_curve_qdotprod;
		double		gmax_curve_qdist;
		double		gmax_curve_qtangacc;
		double		gmax_curve_curvature;
		double		gmax_curve_torsion;
		int			fixedindex;

    //PT method
        void        optPT_setLastQuat( void );
        int         pt_flag_first_time;
        double      pt_curve_dist;
        int         pt_update_ring_index;
        double      pt_init_value;
        int         pt_flag_terminate;

	//--------------------------------------------------//
	/* GRAPH OPTIONS */
	void		initGraph( void );
	void		computeCurveProperties ( void );
	int			gshowQdist;
	int			gshowQdotprod;
	int			gshowQtangacc;
	int			gshowCcurvature;
	int			gshowCtorsion;


    // LOADING MESHVIEW DATA
    Obj         **olist;
    Obj         **tmpobj;
    Obj         ***lineList;
    int         nobjs;                  // index for single lines
    int         mobjs;                  // index for multiple lines                                 
    int         nlines;
    int         flag_multi_lines;
    int         flag_multi_lines_first_time;
    void        loadData                ( const char *FileName );
    void        drawLINESET             ( void );
    void        drawLINE                ( void );
    void        drawLINESET_tubed       ( void );
    void        drawLINESET_QuatMap     ( void );
};

#endif
