#ifndef _STRUCTS_H
#define _STRUCTS_H

///////////////////////////////////////////////////////////////////////////////
//
// $Header:  structs.h$
//
// NAME: none
//
// BASE  CLASSES: none
//  
// PURPOSE: Global include for data structures
//
// AUTHOR: Sidharth Thakur
//
// DATE: Thu Jun 30 16:53:55 EST 2005
//
// COPYRIGHT NOTICE: "Copyright 2005, Trustees of Indiana University, Hanson 
// Visualization Laboratory"
//
//  REVISION HISTORY:
// $Log:      $
///////////////////////////////////////////////////////////////////////////////

#include <assert.h>

#define MAX_POINTS					1000
#define MAX_TEXTURES				10
#define MAXOBJ                      100

#ifndef M_PI
#define	M_PI						3.14159265358979323846
#endif

#define FALSE						0
#define TRUE						1

#define BUFSIZE						512
#define MAX_TEXTURES				10

#define ON  TRUE
#define OFF FALSE
// OPERATIONS
#define sign(A)                 ((A) < 0.0 ? -1.0 : ((A) > 0.0 ? 1.0 : 0.0))
#define Min(A,B)                ((A) <= (B) ? (A) : (B))
#define Max(A,B)                ((A) >= (B) ? (A) : (B))
#define Inbetween(A,X,B)        ((Min(A,B) <= (X)) && ((X) <= Max(A,B)))
#define zap(x)					{ assert(x != NULL); delete(x); x = NULL; }

/* SPLINE TYPE */
enum {
	CATROM,
	BEZIER,
	BSPLINE,
	NOSPLINE
};

/* MATH OBJECT TYPE */
enum { 
	MATHOBJ = 200,
	OBJKNOT, 
	OBJCURVE, 
    OBJCURVELIST,
	OBJSPLINE,
	OBJSURFACE,
	OBJBELT
};

enum {
	PARALLELTRANSPORT = 300,
	FRENETSERRET,
	GEODESICREFERENCE,
	NOFRAME,
    REFLECTEDQUATLIST,
	OPTIMIZATION
};

/* Window IDs */
enum {
	MAIN_WINDOW = 400,
	AUXY_WINDOW_1,
	AUXY_WINDOW_2
};

/* Texture codes */
enum {
	CODE_SOLID_COLOR = 500,
	CODE_CURVATURE,
	CODE_TORSION,
	CODE_TWO_SIDE,
	CODE_DEPTH,
    CODE_QUAT_TURN_ANGLE,
    CODE_QUAT_DOTPROD
};



/* QUATERNION COMPONENTS */
enum{
	SHOWXYZ,
	SHOWWYZ,
	SHOWWXY,
	SHOWWXZ
};

enum { 
	BELT360,
	BELT720
};

//MESHVIEW DECLARATIONS
enum { NOCOLOR = 650, UNIFORM1, UNIFORM2 };
enum { FORMAT_OFF = 100,  MESH, LIST, FRAMES, DOT, LINE, LINE_STRIP, LINE_LOOP, LINE_LIST };
enum { FLAT = 150, SMOOTH };
enum { NO_COLOR, VERT_FRONT, VERT_BACK, FACE_FRONT, FACE_BACK, DEPTH_C, TEX_C };
enum { SAVE_PPM = 400, SAVE_SETTING, LOAD_DATA, LOAD_NEW_DATA, LOAD_SETTING, LOAD_PALETTE, LOAD_TEXTURE };
enum { NEW = 450, ADD };
enum { SAME = 500, DIFF, DEPTH, TEXTURE };
enum { SD_OFF = 520, SD_PLUS, SD_MINUS };
enum { COLORDFT = 550, COLOROBJ, COLORFACE, COLORVERT };
enum { PERSPECTIVE = 600, ORTHOGONAL, POLAR };
enum { SETTING1_DIALOG = 700, SETTING2_DIALOG, PICKER_DIALOG, FRAMES_DIALOG, HELP_DIALOG, ABOUT_DIALOG, MOUSE3D_DIALOG, QUATROT_DIALOG};

typedef struct { double d[3]; } Point3;
typedef struct { double d[4]; } Point4;
typedef struct { double d[3]; } Vector3;
typedef struct { double d[4]; } Vector4;

typedef struct { double d[3]; } Texture;
typedef struct { float f[4]; } Color4; 

typedef struct Vertex Vertex;
typedef struct Face Face;
typedef struct Obj Obj;


struct Vertex{
    int index;            /* index of this vertex in current obj */
    Point3 p3;            /* 3d coordinates */
    int depth;            /* 4d depth between 0 and 255 */
    Point4 p4;            /* 4d coordinates */
    /*Point5 p5;             5d coordinates */
    /*Point6 p6;             6d coordinates */
    /*PointN pN;             Nd coordinates */
    Vector3 normal ;      /* 3d vertex normal */
    Color4 cfront, cback; /* RGBA colors */
    Texture tex;          /* Texture coordinates */
    int Nfaces;           /* number of neighbouring faces */
    Face ** faces;        /* list of pointers to neighbouring faces */
    int		color;				//similar to depth
};
struct Face {
  int index;            /* index of this face in current obj */
  int nvert;            /* # of vertices in the face (4 in MESH) */
  Vertex ** vertices;   /* pointers to the vertices */  
  Vector3 normal;       /* 3d plane normal */
  Color4 cfront, cback; /* RGBA colors */
};
typedef struct {
    int width, height;        // Width and height of the texture
    unsigned char *buffer;    // Pointer to memory that holds the texture information.
} Image;

struct Obj {
  int objectID;
  int gformat;       /* FORMAT_OFF, MESH, LIST, FRAMES, DOT, LINES etc.*/
  int cformat;       /* UNIFORM (0, 1 or 2 colors per obj) */ 
  int texture_flag;  /* Does user supply texture coordinates for vertices */
  int vcolor;        /* number of colors supplied for each vertex, 0, 1 or 2 */
  int fcolor;        /* number of colors supplied for each face, 0, 1 or 2 */
  int nvertices, nfaces;
  int nlines;        /* @sid number of lines */
  int nu, nv;        /* sizes of mesh, useful only when format_flag = MESH */
  int u_flag;        /* part of MESH format */
  int v_flag;        /* part of MESH format */
  int size_flag;     /* valid in POINTS/ LINES formats */
  float size;        

  Vertex *vlist;
  Face *flist;       /* in LINES format holds lines */
  double scale;      /* scale and center are used to fit object */
  //Point4 center;     /* into identity sphere */
  Color4 cfront, cback;

  int nframes;       /* data for frame mode */
  //Obj **frames;

};
//struct Face {
//    int       index;             // index of this face in current obj 
//    int       nvert;             // # of vertices in the face (4 in MESH) 
//    Vertex ** vertices;          // pointers to the vertices 
//    Vector3   normal;            // 3d plane normal 
//    Color4    cfront, cback;     // RGBA colors 
//};
//struct Obj {
// int    gformat;       /* FORMAT_OFF, MESH, LIST, FRAMES, DOT, LINES, GLM_OBJ etc.*/
// int    cformat;       /* UNIFORM (0, 1 or 2 colors per obj) */ 
// int    texture_flag;  /* Does user supply texture coordinates for vertices */
// int    vcolor;        /* number of colors supplied for each vertex, 0, 1 or 2 */
// int    fcolor;        /* number of colors supplied for each face, 0, 1 or 2 */
// int    nvertices, nfaces;
// int    nu, nv;        /* sizes of mesh, useful only when format_flag = MESH */
// int    u_flag;        /* part of MESH format */
// int    v_flag;        /* part of MESH format */
// int    size_flag;     /* valid in POINTS/ LINES formats */
// float   size;        
//
// Vertex *vlist;
// Face	*flist;   /* in LINES format holds lines */
// double   scale;   /* scale and center are used to fit object */
// Point4   center;   /* into identity sphere */
// Color4   cfront, cback; /* color */
//
// int    nframes;       /* data for frame mode */
// Obj    **frames;
//
// //GLM object part
// //GLMmodel  *glmObj;
//
// //Pointer to curve's tube
// ThickenedCurve **curveTube;
//
// //for alpha blending 
// //the type of a blending 
// //tree is a BSPTree
// //int rebuild_blending_tree_flag;
// //void *blending_tree;
//};
//
//
//
//typedef struct node{
// double v;
// struct node *next;
//} node;

//typedef struct {
//    int width, height;        // Width and height of the texture
//    unsigned char *buffer;    // Pointer to memory that holds the texture information.
//} Image;

//typedef struct {
// node *begin;
// node *end;
// int count;
//} queue;
//

/* Lighting structures */
static const GLfloat no_mat[]						= {0.0f,0.0f,0.0f,1.0f};
static const GLfloat mat_ambient[]					= {0.7f,0.7f,0.7f,1.0f};
static const GLfloat mat_ambient_color[]			= {0.8f,0.8f,0.2f,1.0f};
static const GLfloat mat_diffuse[]					= {0.1f,0.5f,0.8f,1.0f};
static const GLfloat mat_specular[]					= {1.0f,1.0f,1.0f,1.0f};
static const GLfloat no_shininess[]					= {0.0f};
static const GLfloat low_shininess[]				= {5.0f};
static const GLfloat high_shininess[]				= {100.0f};
static const GLfloat mat_emission[]					= {0.3f,0.2f,0.2f,0.0f};

#endif //_STRUCTS_H
